"""Environment variables used by OsmanConfig."""
OSMAN_ENVIRONMENT_VARS = [
    "OPENSEARCH_HOST",
    "OPENSEARCH_PORT",
    "OPENSEARCH_SSL_ENABLED",
    "OPENSEARCH_USER",
    "OPENSEARCH_SECRET",
    "AUTH_METHOD",
    "AWS_ACCESS_KEY_ID",
    "AWS_SECRET_ACCESS_KEY",
    "AWS_REGION",
    "AWS_SERVICE",
]
